import struct
import io
import collections
import tempfile
import shutil
import abc

from array import array

from pytag import utils


PacketInfo = collections.namedtuple('PacketInfo', ['size', 'complete'])


class OggPage:
    """
    """

    header_struct = struct.Struct("< 4s B B q I I I B")

    def __init__(self, fileobj):
        """Inizializates a OggPage

        :param fileobj: Stream to read from.
        :type fileobj: ``io.RawIOBase``
        """

        self.fileobj = fileobj
        self.segment_table_index = 0

        self._unpack(fileobj.read(27))
        self.segment_table = array('B', fileobj.read(self.page_segments))
        self.packet_reader = None

    def next_page(self):
        """Iterates to next page.
        """

        # If page has not readed segments, seek them
        self.fileobj.seek(sum(self.segment_table[self.segment_table_index:]),
                          io.SEEK_CUR)
        self.__init__(self.fileobj)
        return self

    def rest_of_pages(self):
        """Iterator over still not readed pages.
        """

        while not self.is_last_page():
            yield self.next_page()

    def as_bytes(self, update_crc=False):
        """Get the complete page as bytes.

        :paramer update_crc: Flag to know if recalculate the CRC.
        :type update_crc: ``boolean``
        :returns: Ogg page
        :rtype: :py:class:`array.array` of bytes, typecode = 'B'
        """

        if update_crc:
            self.crc = 0

        page = array('B')
        page.extend(self._pack())
        page.extend(self.segment_table)
        page.extend(self.fileobj.read(sum(self.segment_table)))
        self.segment_table_index = len(self.segment_table)

        if update_crc:
            crc = utils.crc32(page)
            page[22:26] = array('B', crc)

        return page

    def is_last_page(self):
        """Check if the page is the last in a logical bitstream.
        :returns: True if is the last one, False if not.
        :rtype: ``boolean``
        """

        return self.header_type == 4

    def _pack(self):
        """Packs the page header.
        :returns: A bytes object containing the page header.
        :rtype: ``bytes``
        """

        return self.header_struct.pack(self.oggs, self.version,
                                       self.header_type, self.granule_position,
                                       self.serial, self.number, self.crc,
                                       self.page_segments)

    def _unpack(self, buf):
        """Unpacks a bytes buffer to set the page header values.
        """

        (self.oggs, self.version, self.header_type, self.granule_position,
         self.serial, self.number, self.crc, self.page_segments) = (
             self.header_struct.unpack(buf))

    def get_packet_info(self):
        """Gets the size of the next packet (or partial packet) in the current
        page. This size can be smaller than the full packet because a packet
        can we splited in severa pages. Also check it the packet finish in the
        current page.  If page has no more bytes to read, this function
        iterates to the next page.
        The idea is use this function as callback in PacketReader, this way,
        when a packet is readed, the reader doesn't need to worry about how the
        packet is saved inside an ogg stream.

        :returns: Next packet size and if the packet finish in this page.
        :rtype: :py:class:`collections.namedtuple` of type ``PacketInfo``
        """

        # Check if current page has more segments to read
        if len(self.segment_table) == self.segment_table_index:
            self.next_page()

        total_size = 0
        for size in self.segment_table[self.segment_table_index:]:
            self.segment_table_index += 1
            total_size += size
            if size is not 255:
                return PacketInfo(size=total_size, complete=True)

        return PacketInfo(size=total_size, complete=False)

    def get_packet_reader(self):
        """Get a packet reader for the current page.

        :returns: A packet reader over the same stream used by this OggPage
        :rtype: ``PacketReader``
        """

        if not self.packet_reader:
            self.packet_reader = PacketReader(self.fileobj,
                                              self.get_packet_info)
        return self.packet_reader

    def __str__(self):   # pragma: no cover
        return ('Magic:  {o.oggs}, '
                'Version: {o.version}, '
                'Header type: {o.header_type}, '
                'Granule position: {o.granule_position}, '
                'Serial Nr: {o.serial}, '
                'Position: {o.number}, '
                'CRC: {o.crc}, '
                'Page segments: {o.page_segments}'
                .format(o=self))


class PacketReader:
    """My class doc."""

    def __init__(self, fileobj, get_packet_info_callback):
        """Method doc"""

        self.fileobj = fileobj
        self.get_packet_info_callback = get_packet_info_callback
        self.position = 0

    def read(self, _n=-1):
        """Read up to n bytes from the current packet in the stream and return
        them. If n is unspecified or -1, read and return all the bytes until
        the packet end.
        """

        if self.position == 0:
            self.limit, self.complete = self.get_packet_info_callback()

        # Read complete packet
        if _n == -1:
            partial_n = self.limit - self.position
            ret = self.fileobj.read(partial_n)
            while not self.complete:
                self.position = 0
                self.limit, self.complete = self.get_packet_info_callback()
                ret += self.fileobj.read(self.limit)

        # Read n bytes from packet
        else:
            ret = b''
            while _n != 0:
                if _n + self.position <= self.limit:
                    partial_n = _n
                else:
                    partial_n = self.limit - self.position
                _n -= partial_n
                self.position += partial_n
                ret += self.fileobj.read(partial_n)
                if _n != 0:
                    self.position = 0
                    self.limit, self.complete = self.get_packet_info_callback()

        return ret


class OggReader(metaclass=abc.ABCMeta):

    def __init__(self, path):
        self.path = path

    def get_tags(self):
        with open(self.path, 'rb') as input_file:
            current_page = OggPage(input_file)
            for i in range(self.comments_page_position()):
                current_page.next_page()
            tags = self.process_comments(current_page.get_packet_reader())

        return tags

    @abc.abstractmethod
    def comments_page_position(self):
        """Returns the page number where the comments start."""

    @abc.abstractmethod
    def process_comments(self, packet):
        """Returns the comments."""


class Ogg(OggReader):

    def __init__(self, path):

        self.output_page_nr = 0
        self.page_out = array('B')
        super().__init__(path)

    @abc.abstractmethod
    def packets_after_comments(self):
        """Returns the number of packets in the same page after the comments
        packet
        """

    def write_tags(self, comments):
        """Write the tags to a new file, if no path is provided, the original
        file is overwrited
        """

        with tempfile.NamedTemporaryFile('wb',
                                         delete=False) as self.output_file,\
                open(self.path, 'rb') as input_file:

            # First page, get serial number and write
            current_page = OggPage(input_file)
            self.serial = current_page.serial
            self.output_file.write(current_page.as_bytes())

            # Second page
            current_page.next_page()

            # Get a packet reader
            packet_reader = current_page.get_packet_reader()

            # Ignore old comments, advance to next packet
            packet_reader.read()

            # Generate and write new comments
            codec_packet = self.generate_comments(comments)
            self._to_page(codec_packet)

            # Read setup header
            for i in range(self.packets_after_comments()):
                codec_packet = packet_reader.read()
                self._to_page(io.BytesIO(codec_packet), force_page_end=True)

            new_pages = self.output_page_nr - current_page.number
            # We need to increment the page secuence number for all pages
            if new_pages > 0:
                for page in current_page.rest_of_pages():
                    page.number += new_pages
                    self.output_file.write(page.as_bytes(update_crc=True))
            else:
                self.output_file.write(input_file.read())

        shutil.move(self.output_file.name, self.path)

    def _to_page(self, packet, p_type=0, force_page_end=False):

        if len(self.page_out) is 0:
            self.page_out.extend(b'OggS')
            self.page_out.extend([0,                       # Version
                                  p_type,                  # Header type
                                  0, 0, 0, 0, 0, 0, 0, 0,  # Granular position
                                  ])

            self.page_out.extend(utils.int_struct.pack(self.serial))  # Ser.Nr
            self.output_page_nr += 1
            self.page_out.extend(utils.int_struct.pack(self.output_page_nr))

            self.page_out.extend([0, 0, 0, 0,  # CRC
                                  0])        # Nr page segments

            self.s_table = array('B')
            self.segments = array('B')

        for segment in utils.read_in_chunks(packet, 255):
            self.s_table.append(len(segment))
            self.segments.extend(segment)
            if len(self.s_table) == 255:
                # Page is full, write and continue in new page
                self._write_page()
                self.page_out = array('B')
                self._to_page(packet, p_type=1)

        # Finish this page
        if force_page_end:
            self._write_page()
            self.page_out = array('B')

    def _write_page(self):
        self.page_out[-1] = len(self.s_table)
        crc = utils.crc32(self.page_out, self.s_table, self.segments)
        self.page_out[22:26] = array('B', crc)
        self.output_file.write(self.page_out)
        self.output_file.write(self.s_table)
        self.output_file.write(self.segments)



import struct
import io
from array import array

from pytag import utils
from pytag.structures import CaseInsensitiveDict
from pytag.constants import VENDOR_NAME


class VorbisComment:
    """Base class to read/write vorbis comments as defined at:
    http://www.xiph.org/vorbis/doc/Vorbis_I_spec.html
    """

    vendor_name = VENDOR_NAME.encode()
    framing_bit = True

    def process_comments(self, packet):
        """Reads the comments.

        :param packet: Object to read from, has a read method.
        :type packet: ``pytag.containers.PacketReader``
        :returns: A ``dict``-like object with all the comments.
        :rtype: ``pytag.structures.CaseInsensitiveDict``
        """

        # Signature is not used
        packet.read(self.signature_struct.size)

        (vendor_lenght,) = utils.int_struct.unpack(packet.read(4))
        packet.read(vendor_lenght)

        (user_comment_list_length,) = utils.int_struct.unpack(packet.read(4))

        comments = CaseInsensitiveDict()
        for i in range(user_comment_list_length):
            (length,) = utils.int_struct.unpack(packet.read(4))
            comment = packet.read(length).decode('utf-8')
            comments.update((comment.split('='),))

        return comments

    def generate_comments(self, comments):
        comments = CaseInsensitiveDict(comments)

        # Packet type + vorbis magic
        packet = array('B', self.signature)

        packet.extend(utils.int_struct.pack(len(self.vendor_name)))
        packet.extend(self.vendor_name)

        packet.extend(utils.int_struct.pack(len(comments)))

        for k, v in comments.items():
            comment = '{}={}'.format(k, v).encode()
            packet.extend(utils.int_struct.pack(len(comment)))
            packet.extend(comment)

        if self.framing_bit:
            packet.append(1)

        return io.BytesIO(packet)


class Vorbis(VorbisComment):

    signature = (0x3, 0x76, 0x6f, 0x72, 0x62, 0x69, 0x73)
    signature_struct = struct.Struct('< B 6s')

    def comments_page_position(self):
        return 1

    def packets_after_comments(self):
        return 1


class Opus(VorbisComment):      # pragma: no cover

    signature = (0x4f, 0x70, 0x75, 0x73, 0x54, 0x61, 0x67, 0x73)
    signature_struct = struct.Struct('< 8s')
    framing_bit = False

    def comments_page_position(self):
        return 1

    def packets_after_comments(self):
        return 0



import collections

from pytag.constants import FIELD_NAMES


class CaseInsensitiveDict(collections.MutableMapping):
    """A case-insensitive :py:class:`dict`-like object.

    Implements all methods and operations of
    :py:class:`collections.abc.MutableMapping` as well as dict's
    :py:meth:`dict.copy`.

    All keys are expected to be strings. The structure always set the key to
    lower case.

    ::

        cid = CaseInsensitiveDict()
        cid['Key'] = 'value'
        cid['KEY'] == 'value'  # True
        list(cid) == ['key']   # True

    If the constructor, update, or equality comparison operations are given
    keys that have equal :py:meth:`str.lower`, the behavior is undefined.
    """

    def __init__(self, data=None, **kwargs):
        self._store = dict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key, value):
        self._store[key.lower()] = value

    def __getitem__(self, key):
        return self._store[key.lower()]

    def __delitem__(self, key):
        del self._store[key.lower()]

    def __iter__(self):
        return (key for key in self._store)

    def __len__(self):
        return len(self._store)

    def __eq__(self, other):
        return other == self._store

    def copy(self):
        return CaseInsensitiveDict(self._store)

    def __repr__(self):    # pragma: no cover
        return '{}({})'.format(self.__class__.__name__, self._store)


class PytagDict(CaseInsensitiveDict):
    """ A case-insensitive :py:class:`dict`-like object where only the values
    defines in :py:data:`pytag.constants.FIELD_NAMES` constant are allowed as
    keys. If a key is not valid, is ignored without any warning.
    """

    def __setitem__(self, key, value):
        key = key.lower()
        if key in FIELD_NAMES:
            self._store[key] = value



import collections
import struct
import io
import tempfile
import shutil
from array import array

from pytag import utils
from pytag.containers import OggReader, Ogg
from pytag.codecs import Vorbis, Opus
from pytag.constants import (ID3_ENCODINGS, ID3_GENRES, FIELD_NAMES,
                             TAG_ID3_V22, TAG_ID3_V23, TAG_ID3_V24)


class OggVorbisReader(Vorbis, OggReader):
    pass


class OggVorbis(Vorbis, Ogg):
    pass


class OggOpusReader(Opus, OggReader):
    pass


class OggOpus(Opus, Ogg):
    pass


Id3Frame = collections.namedtuple('Id3Frame', ['comment', 'size'])


class Mp3Reader:

    def __init__(self, path):
        self.path = path

    def get_tags(self):

        tags = {}
        with open(self.path, 'rb') as self.input_file:

            if self._has_id3v2_tags():
                tags = self._read_id3v2_tags()
            elif self._has_id3v1_tags():
                tags = self._read_id3v1_tags()

        return tags

    def _has_id3v1_tags(self):
        self.input_file.seek(0, io.SEEK_END)
        if self.input_file.tell() > 128:
            self.input_file.seek(-128, io.SEEK_END)
            (tag, ) = struct.unpack('> 3s', self.input_file.read(3))
            if tag == b'TAG':
                return True

        return False

    def _has_id3v2_tags(self):
        (id3, ) = struct.unpack('> 3s', self.input_file.read(3))
        return id3 == b'ID3'

    def _read_id3v1_tags(self):
        tags = {}

        title = self._remove_padding(self.input_file.read(30))
        if title:
            tags['title'] = title

        artist = self._remove_padding(self.input_file.read(30))
        if artist:
            tags['artist'] = artist

        album = self._remove_padding(self.input_file.read(30))
        if album:
            tags['album'] = album

        date = self._remove_padding(self.input_file.read(4))
        if date:
            tags['date'] = date

        self.input_file.seek(29, io.SEEK_CUR)  # Ignore comments

        tracknumber = int.from_bytes(self.input_file.read(1), byteorder='big')
        if tracknumber:
            tags['tracknumber'] = tracknumber

        genre = int.from_bytes(self.input_file.read(1), byteorder='big')
        if genre in ID3_GENRES:
            tags['genre'] = ID3_GENRES[genre]

        return tags

    def _remove_padding(self, text):
        try:
            return text[:text.index(b'\x00')].decode()
        except ValueError:
            return text.decode()

    def _read_id3v2_tags(self):
        (mayor, minor) = struct.unpack('> B B', self.input_file.read(2))
        self.input_file.seek(1, io.SEEK_CUR)
        size = utils.decode_bitwise_int(struct.unpack('> B B B B',
                                                      self.input_file.read(4)))

        if mayor == 2:
            read_frame = self._read_id3v22_frame
        elif mayor == 3:
            read_frame = self._read_id3v23_frame
        elif mayor == 4:
            read_frame = self._read_id3v24_frame
        else:   # pragma: no cover
            raise Exception('ID3 version "2.{}" not supported'.format(mayor))

        comments = {}
        while size > 0:
            frame = read_frame()
            size -= frame.size
            if frame.comment:
                comments.update(frame.comment)

        return comments

    def _decode_genre(self, text):
        genres = []
        for code in text.split(')('):
            code = int(code.strip('()'))
            genres.append(ID3_GENRES[code])

        if len(genres) == 1:
            return genres[0]

        return genres

    def _read_id3_generic_frame(self, frame_id, size, id3_type):

        (encoding, ) = struct.unpack('> B', self.input_file.read(1))
        header_size = 6 if id3_type == 2 else 10
        try:
            if id3_type == 2:
                index = TAG_ID3_V22.index(frame_id.decode())
            if id3_type == 3:
                index = TAG_ID3_V23.index(frame_id.decode())
            if id3_type == 4:
                index = TAG_ID3_V24.index(frame_id.decode())

            data = self.input_file.read(size-1)
            data = data.decode(ID3_ENCODINGS[encoding])

            if frame_id == b'TCO' or frame_id == b'TCON':
                if id3_type != 4:
                    data = self._decode_genre(data)

            data = {FIELD_NAMES[index]: data}
            return Id3Frame(data, size + header_size)

        except ValueError:
            self.input_file.seek(size - 1, io.SEEK_CUR)
            return Id3Frame(None, size + header_size)

    def _read_id3v22_frame(self):
        (frame_id, ) = struct.unpack('> 3s ', self.input_file.read(3))
        size = int.from_bytes(self.input_file.read(3), byteorder='big')
        return self._read_id3_generic_frame(frame_id, size, 2)

    def _read_id3v23_frame(self):
        (frame_id, ) = struct.unpack('> 4s ', self.input_file.read(4))
        (size, ) = struct.unpack('> I', self.input_file.read(4))

        self.input_file.seek(2, io.SEEK_CUR)  # Flags, not used

        return self._read_id3_generic_frame(frame_id, size, 3)

    def _read_id3v24_frame(self):
        (frame_id, ) = struct.unpack('> 4s ', self.input_file.read(4))
        size = utils.decode_bitwise_int(struct.unpack('> B B B B',
                                                      self.input_file.read(4)))

        self.input_file.seek(2, io.SEEK_CUR)  # Flags, not used

        return self._read_id3_generic_frame(frame_id, size, 4)


class Mp3(Mp3Reader):

    def write_tags(self, comments):

        with tempfile.NamedTemporaryFile('wb', delete=False) as output_file,\
                open(self.path, 'rb') as self.input_file:

            # Write tags if at least has one supported value
            tags = array('B')
            for key, value in comments.items():
                if key in FIELD_NAMES:

                    frame = array('B',
                                  TAG_ID3_V24[FIELD_NAMES.index(key)].encode())
                    frame.extend([0]*6)  # Size and flags
                    frame.append(3)      # Encoding
                    frame.extend(value.encode())
                    frame[4:8] = utils.encode_bitwise_int(len(frame) - 10)

                    tags.extend(frame)

            if len(tags):
                output_file.write(b'ID3\x04\x00\x00')
                output_file.write(utils.encode_bitwise_int(len(tags)))
                output_file.write(tags)

            # Position input file cursor after id3v2 tags
            if self._has_id3v2_tags():
                self._read_id3v2_tags()
            else:
                self.input_file.seek(0)

            output_file.write(self.input_file.read())

            if self._has_id3v1_tags():
                output_file.seek(-128, io.SEEK_END)
                output_file.truncate()

        shutil.move(output_file.name, self.path)



from os.path import splitext

from pytag.structures import PytagDict
from pytag.constants import FIELD_NAMES
from pytag.formats import OggVorbisReader, OggVorbis, Mp3Reader, Mp3


EXTENSIONS = {'.ogg': (OggVorbisReader, OggVorbis),
              '.mp3': (Mp3Reader, Mp3)
              }


class Tag:
    """Descriptor class.
    """

    def __init__(self, name):
        self.name = name

    def __get__(self, instance, cls):

        if instance is None:  # pragma: nocover
            return self

        try:
            return instance.__dict__[self.name]
        except KeyError:
            tags = instance.get_tags()
            for name in (set(tags) ^ set(FIELD_NAMES)):
                instance.__dict__[name] = None
            return instance.__dict__[self.name]

    def __set__(self, instance, value):
        instance.__dict__[self.name] = value


class MetaAudio(type):
    """Set all the FIELD_NAMES as class descriptors.
    """

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for name in FIELD_NAMES:
            setattr(cls, name, Tag(name))


class AudioReader(metaclass=MetaAudio):
    """High level interface for pytag. Creates a new object if the audio format
    is supported, or returns a :py:exc:`pytag.FormatNotSupportedError` if not.
    """

    _index = 0

    def __init__(self, path):
        try:
            ext = splitext(path)[1].lower()
            self._format = EXTENSIONS[ext][self._index](path)
        except KeyError:
            raise FormatNotSupportedError(
                '"{}" extension is not suppored'.format(ext))

    def get_tags(self):

        tags = PytagDict(self._format.get_tags())
        for name, value in tags.items():
            setattr(self, name, value)
        return tags


class Audio(AudioReader):
    """Extends :py:class:`pytag.AudioReader` and adds a ``write_tags`` method.
    """

    _index = 1

    def write_tags(self, tags):
        self._format.write_tags(PytagDict(tags))


class FormatNotSupportedError(Exception):
    pass



import struct
from array import array

from pytag.constants import CRC_LOOKUP


int_struct = struct.Struct('< I')


def read_in_chunks(fileobj, chunk_size=255):
    """Read a file in chunks."""
    chunk = fileobj.read(chunk_size)
    while chunk:
        # Do stuff with byte.
        yield chunk
        chunk = fileobj.read(chunk_size)


def crc32(*args):

    crc_reg = 0
    for element in args:
        for i, value in enumerate(element):
            crc_reg = ((crc_reg << 8) ^ CRC_LOOKUP[((crc_reg >> 24) & 0xff) ^
                                                   element[i]]) & 0xFFFFFFFF

    a = (crc_reg & 0xff)
    b = ((crc_reg >> 8) & 0xff)
    c = ((crc_reg >> 16) & 0xff)
    d = ((crc_reg >> 24) & 0xff)

    return a, b, c, d


def decode_bitwise_int(tup):
    d, c, b, a = tup
    return ((a | (b << 7)) | (c << 14)) | (d << 21)


def encode_bitwise_int(number):

    r = array('B')
    mask = 127  # 01111111
    r.insert(0, mask & number)
    r.insert(0, ((mask << 7) & number) >> 7)
    r.insert(0, ((mask << 14) & number) >> 14)
    r.insert(0, ((mask << 21) & number) >> 21)
    return r



from distutils.core import setup
import os

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst'), 'rt') as f:
    README = f.read()

setup(name='pytag',
      version='0.1.2',
      author='José Luis Lafuente',
      author_email='jlesquembre@gmail.com',
      description='A library to handle audio metadata',
      long_description=README,
      license='GNU General Public License v3 (GPLv3)',
      url='http://jlesquembre.github.io/pytag/',
      packages=['pytag'],
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Utilities',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.3',
        ],
      keywords=['audio', 'metadata'],
    )
